#   Copyright (C) 2006-2007 Daniel Burrows
#   Copyright (C) 2023 Scott Hansen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

"""An urwid listview-based widget that lets you choose a URL from a list of
URLs."""

import contextlib
import json
import os
from os.path import dirname, exists, expanduser
from sys import platform
import re
import shlex
import subprocess
import sys
from threading import Thread
import webbrowser

import urwid


if platform == 'darwin':
    COPY_COMMANDS = ('pbcopy',)
    COPY_COMMANDS_PRIMARY = ('pbcopy',)
    import selectors
    selectors.DefaultSelector = selectors.SelectSelector
elif 'WAYLAND_DISPLAY' in os.environ:
    COPY_COMMANDS = ('wl-copy',)
    COPY_COMMANDS_PRIMARY = ('wl-copy --primary',)
else:
    COPY_COMMANDS = ("xsel -ib", "xclip -i -selection clipboard")
    COPY_COMMANDS_PRIMARY = ("xsel -i", "xclip -i")

# Text-mode browsers take over the terminal and render to stdout, so they
# must be run in the foreground with the urwid screen suspended.
TERMINAL_BROWSERS = ('elinks', 'links', 'w3m', 'lynx')


def shorten_url(url, cols, shorten):
    """Shorten long URLs to fit on one line.

    """
    cols = ((cols - 6) * .85)  # 6 cols for urlref and don't use while line
    if shorten is False or len(url) < cols:
        return url
    split = int(cols * .5)
    return url[:split] + "..." + url[-split:]


def grp_list(items):
    """Organize list of items [a,2,3,4,a,4,2,a,1, etc...] like:
        [[a,2,3,4], [a,4,2], [a,1]], where 'a' is a urwid.Divider

    """
    grp = []
    res = []
    for item in items:
        if isinstance(item, urwid.Divider):
            res.append(grp)
            grp = [items[0]]
        else:
            grp.append(item)
    res.append(grp)
    return res[1:]


def splittext(text, search, attr):
    """Split a text string by search string and add Urwid display attribute to
    the search term.

    Args: text - string
          search - search string
          attr - attribute string to add

    Returns: urwid markup list ["string", ("default", " mo"), "re string"]
             for search="mo", text="string more string" and attr="default"

    """
    if search:
        pat = re.compile(f"({re.escape(search)})", re.IGNORECASE)
    else:
        return text
    final = pat.split(text)
    final = [(attr, i) if i.lower() == search.lower() else i for i in final]
    return final


class URLChooser:

    def __init__(self, extractedurls, compact=False, reverse=False, nohelp=False, dedupe=False,
                 shorten=True, run="", runsafe="", single=False, pipe=False,
                 genconf=False, width=0, whitespaceoff=False, colors="true"):
        self.conf = expanduser("~/.config/urlscan/config.json")
        self.keys = {'/': self._search_key,
                     '0': self._digits,
                     '1': self._digits,
                     '2': self._digits,
                     '3': self._digits,
                     '4': self._digits,
                     '5': self._digits,
                     '6': self._digits,
                     '7': self._digits,
                     '8': self._digits,
                     '9': self._digits,
                     'a': self._add_url,
                     'C': self._clipboard,
                     'c': self._context,
                     'ctrl l': self._clear_screen,
                     'ctrl f': self._page_down,
                     'ctrl b': self._page_up,
                     'd': self._del_url,
                     'f1': self._help_menu,
                     'G': self._bottom,
                     'g': self._top,
                     'j': self._down,
                     'k': self._up,
                     'J': self._next,
                     'K': self._previous,
                     'P': self._clipboard_pri,
                     'l': self._link_handler,
                     'o': self._open_queue,
                     'O': self._open_queue_win,
                     'p': self._palette,
                     'Q': self._quit,
                     'q': self._quit,
                     'R': self._reverse,
                     'S': self._all_shorten,
                     's': self._shorten,
                     'u': self._all_escape
                     }
        # Set urwid color mode based on the colors parameter
        self.color = {
            "true": 2**24,
            "256": 256,
            "88": 88,
            "16": 16,
            "8": 8,
            "mono": 1
        }.get(colors, 16)
        self.palettes = {}
        # Default color palette
        default = [('header', 'white', 'dark blue', 'standout', '#ffffff', '#0000aa'),
                   ('footer', 'white', 'dark red', 'standout', '#ffffff', '#aa0000'),
                   ('search', 'white', 'dark green', 'standout', '#ffffff', '#00aa00'),
                   ('msgtext', '', '', '', '', ''),
                   ('msgtext:ellipses', 'light gray', 'black', '', '#aaaaaa', '#000000'),
                   ('urlref:number:braces', 'light gray', 'black', '', '#aaaaaa', '#000000'),
                   ('urlref:number', 'yellow', 'black', 'standout', '#ffff00', '#000000'),
                   ('urlref:url', 'white', 'black', 'standout', '#ffffff', '#000000'),
                   ('url:sel', 'white', 'dark blue', 'bold', '#ffffff', '#0000aa')]
        # Default black & white palette
        blw = [('header', 'black', 'light gray', 'standout', '#000000', '#aaaaaa'),
               ('footer', 'black', 'light gray', 'standout', '#000000', '#aaaaaa'),
               ('search', 'black', 'light gray', 'standout', '#000000', '#aaaaaa'),
               ('msgtext', '', '', '', '', ''),
               ('msgtext:ellipses', 'white', 'black', '', '#ffffff', '#000000'),
               ('urlref:number:braces', 'white', 'black', '', '#ffffff', '#000000'),
               ('urlref:number', 'white', 'black', 'standout', '#ffffff', '#000000'),
               ('urlref:url', 'white', 'black', 'standout', '#ffffff', '#000000'),
               ('url:sel', 'black', 'light gray', 'bold', '#000000', '#aaaaaa')]
        # Default catppuccin palette
        ctp = [("header", "white", "dark blue", "standout", "#CDD6F4", "#89B4FA"),
               ("footer", "white", "dark red", "standout", "#CDD6F4", "#F38BA8"),
               ("search", "white", "dark green", "standout", "#CDD6F4", "#A6E3A1"),
               ("msgtext", "", "", "", "#CDD6F4", "#1E1E2E"),
               ("msgtext:ellipses", "light gray", "black", "", "#B4BEFE", "#1E1E2E"),
               ("urlref:number:braces", "light gray", "black", "", "#B4BEFE", "#1E1E2E"),
               ("urlref:number", "yellow", "black", "standout", "#F9E2AF", "#1E1E2E"),
               ("urlref:url", "white", "black", "standout", "#CBA6F7", "#1E1E2E"),
               ("url:sel", "white", "dark blue", "bold", "#F5E0DC", "#313244")]

        if genconf is True:
            # Add default palettes for config generation
            self.palettes.update([("default", default), ("bw", blw), ("catppuccin", ctp)])
            self._config_create()

        # Flag to track if we found palettes in config.json
        config_has_palettes = False

        try:
            with open(self.conf, 'r', encoding=sys.getdefaultencoding()) as conf_file:
                data = json.load(conf_file)
                try:
                    if 'palettes' in data and data['palettes']:
                        # If config has palettes, use only those
                        config_has_palettes = True
                        for pal_name, pal in data['palettes'].items():
                            self.palettes.update([(pal_name, [tuple(i) for i in pal])])
                except KeyError:
                    pass
                try:
                    items = data['keys'].items()
                    for key, value in items:
                        if value:
                            if value == "open_url":
                                urwid.Button._command_map._command[key] = 'activate'
                            value = getattr(self, f"_{value}")
                        else:
                            del self.keys[key]
                            continue
                        self.keys.update([(key, value)])
                except KeyError:
                    pass
        except FileNotFoundError:
            pass

        # If no palettes were found in config, use the default hardcoded ones
        if not config_has_palettes:
            self.palettes.update([("default", default), ("bw", blw), ("catppuccin", ctp)])
        try:
            subprocess.run(['xdg-open'], check=False, stdout=subprocess.DEVNULL)
            self.xdg = True
        except OSError:
            self.xdg = False
        self.shorten = shorten
        self.compact = compact
        self.queue = []
        self.run = run
        self.runsafe = runsafe
        self.single = single
        self.pipe = pipe
        self.search = False
        self.search_string = ""
        self.no_matches = False
        self.enter = False
        self.term_width, _ = urwid.raw_display.Screen().get_cols_rows()
        self.width = min(self.term_width, width or self.term_width)
        self.whitespaceoff = whitespaceoff
        self.activate_keys = [i for i, j in urwid.Button._command_map._command.items()
                              if j == 'activate']
        self.items, self.urls = self.process_urls(extractedurls,
                                                  dedupe=dedupe,
                                                  shorten=self.shorten)
        # Original version of all items
        self.items_orig = self.items
        # Store items grouped into sections
        self.items_org = grp_list(self.items)
        # Store 'compact' mode items
        self.items_com = [i for i in self.items if
                          isinstance(i, urwid.Columns) is True]
        if self.compact is True:
            self.items, self.items_com = self.items_com, self.items
        self.urls_unesc = [i.replace('\\', '') for i in self.urls]
        self.unesc = False
        listbox = urwid.ListBox(self.items)
        self.header = (":: F1 - help/keybindings :: "
                       "q - quit :: "
                       "/ - search :: "
                       "URL opening mode - {} :: "
                       "Queue - {}")
        self.link_open_modes = ["Web Browser", "Xdg-Open"] if self.xdg is True else ["Web Browser"]
        if self.runsafe:
            self.link_open_modes.insert(0, self.runsafe)
        elif self.run:
            self.link_open_modes.insert(0, self.run)
        self.nohelp = nohelp
        if nohelp is False:
            self.headerwid = urwid.AttrMap(urwid.Text(
                self.header.format(self.link_open_modes[0], len(self.queue))), 'header')
        else:
            self.headerwid = None
        self.top = urwid.Frame(listbox, self.headerwid)
        self.pad = self.term_width - self.width
        self.top = urwid.Padding(self.top, left=0, right=self.pad)
        if self.urls:
            self.top.base_widget.body.focus_position = \
                (2 if self.compact is False else 0)
        if reverse is True:
            self._reverse()
        self.tui = urwid.raw_display.Screen()
        self.palette_names = list(self.palettes.keys())
        self.palette_idx = 0
        self.number = ""
        self.help_menu = False

    def main(self):
        """Urwid main event loop

        """
        self.loop = urwid.MainLoop(self.top, self.palettes[self.palette_names[0]], screen=self.tui,
                                   handle_mouse=False, input_filter=self.handle_keys,
                                   unhandled_input=self.unhandled)
        self.loop.screen.set_terminal_properties(self.color)
        self.loop.run()

    @property
    def size(self):
        _, rows = self.tui.get_cols_rows()
        return (self.width, rows)

    def handle_keys(self, keys, raw):
        """Handle widget default keys

            - 'Enter' or 'space' to load URL
            - 'Enter' to end search mode
            - add 'space' to search string in search mode
            - Workaround some small positioning bugs

        """
        for j, k in enumerate(keys):
            if self.search is True:
                text = f"Search: {self.search_string}"
                if k == 'enter':
                    # Catch 'enter' key to prevent opening URL in mkbrowseto
                    self.enter = True
                    if not self.items:
                        self.search = False
                        self.enter = False
                    if self.search_string:
                        footer = 'search'
                    else:
                        footer = 'default'
                        text = ""
                    footerwid = urwid.AttrMap(urwid.Text(text), footer)
                    self.top.base_widget.footer = footerwid
                elif k in self.activate_keys:
                    self.search_string += k
                    self._search()
                elif k == 'backspace':
                    self.search_string = self.search_string[:-1]
                    self._search()
            elif k in self.activate_keys and \
                    self.urls and \
                    self.search is False and \
                    self.help_menu is False:
                self._open_url()
            elif self.help_menu is True:
                self._help_menu()
                return []
            if k == 'up':
                # Works around bug where the up arrow goes higher than the top list
                # item and unintentionally triggers context and palette switches.
                # Remaps 'up' to 'k'
                keys[j] = 'k'
            if k == 'home':
                # Remap 'home' to 'g'. Works around small bug where 'home' takes the cursor
                # above the top list item.
                keys[j] = 'g'
        # filter backspace out before the widget, it has a weird interaction
        return [i for i in keys if i != 'backspace']

    def unhandled(self, key):
        """Handle other keyboard actions not handled by the ListBox widget.

        """
        self.key = key
        if self.search is True:
            if self.enter is False and self.no_matches is False:
                if len(key) == 1 and key.isprintable():
                    self.search_string += key
                self._search()
            elif self.enter is True and not self.search_string:
                self.search = False
                self.enter = False
            return
        if not self.urls and key not in "Qq":
            return  # No other actions are useful with no URLs
        if self.help_menu is False:
            try:
                self.keys[key]()
            except KeyError:
                pass

    def _quit(self):
        """q/Q"""
        raise urwid.ExitMainLoop()

    def _open_url(self):
        """<Enter> or <space>"""
        load_text = "Loading URL..." if self.link_open_modes[0] != (self.run or self.runsafe) \
            else f"Executing: {self.run or self.runsafe}"
        if os.environ.get('BROWSER') not in TERMINAL_BROWSERS:
            self._footer_display(load_text, 5)

    def _background_queue(self, mode):
        """Open URLs in background"""
        for url in self.queue:
            self.mkbrowseto(url, mode=mode, background=True)()
        self.draw_screen()

    def _queue(self, mode=2):
        """Open all URLs in queue

            Args: mode - 2 for new tab, 1 for new window

        """
        load_text = "Loading URLs in queue..." \
            if self.link_open_modes[0] != (self.run or self.runsafe) \
            else f"Executing: {self.run or self.runsafe}"
        if os.environ.get('BROWSER') in TERMINAL_BROWSERS:
            self._footer_display("Opening multiple links not support in text browsers", 5)
            return
        self._footer_display(load_text, 5)
        thr = Thread(target=self._background_queue, args=(mode,))
        thr.start()
        self.queue = []
        self.headerwid = urwid.AttrMap(urwid.Text(
            self.header.format(self.link_open_modes[0], len(self.queue))), 'header')
        self.top.base_widget.header = self.headerwid

    def _open_queue(self):
        """o (new tab)"""
        if self.queue:
            self._queue()

    def _open_queue_win(self):
        """O (new window)"""
        if self.queue:
            self._queue(1)

    def _add_url(self):
        """a"""
        fpo = self.top.base_widget.body.focus_position
        url_idx = len([i for i in self.items[:fpo + 1]
                       if isinstance(i, urwid.Columns)]) - 1
        if self.compact is False and fpo <= 1:
            return
        self.queue.append(self.urls[url_idx])
        self.queue = list(set(self.queue))
        self.headerwid = urwid.AttrMap(urwid.Text(
            self.header.format(self.link_open_modes[0], len(self.queue))), 'header')
        self.top.base_widget.header = self.headerwid
        label = self.items[fpo][1].label
        if not label.startswith("* "):
            self.items[fpo][1].set_label(f"* {label}")

    def _del_url(self):
        """d"""
        fpo = self.top.base_widget.body.focus_position
        url_idx = len([i for i in self.items[:fpo + 1]
                       if isinstance(i, urwid.Columns)]) - 1
        if self.compact is False and fpo <= 1:
            return
        try:
            self.queue.remove(self.urls[url_idx])
            self.headerwid = urwid.AttrMap(urwid.Text(
                self.header.format(self.link_open_modes[0], len(self.queue))), 'header')
            self.top.base_widget.header = self.headerwid
            label = self.items[fpo][1].label
            if label.startswith("* "):
                self.items[fpo][1].set_label(label.lstrip("* "))
        except ValueError:
            pass

    def _help_menu(self):
        """F1"""
        if self.help_menu is False:
            self.focus_pos_saved = self.top.base_widget.body.focus_position
            help_men = "\n".join([f"{i} - {j.__name__.strip('_')}"
                                  for i, j in self.keys.items() if j.__name__ !=
                                  '_digits'])
            help_men = "KEYBINDINGS\n" + help_men + "\n<0-9> - Jump to item"
            docs = ("OPTIONS\n"
                    "add_url       -- add URL to queue\n"
                    "all_escape    -- toggle unescape all URLs\n"
                    "all_shorten   -- toggle shorten all URLs\n"
                    "bottom        -- move cursor to last item\n"
                    "clear_screen  -- redraw screen\n"
                    "clipboard     -- copy highlighted URL to clipboard\n"
                    "                 using xsel/xclip\n"
                    "clipboard_pri -- copy highlighted URL to primary\n"
                    "                 selection using xsel/xclip\n"
                    "config_create -- create ~/.config/urlscan/config.json\n"
                    "context       -- show/hide context\n"
                    "del_url       -- delete URL from queue\n"
                    "down          -- cursor down\n"
                    "help_menu     -- show/hide help menu\n"
                    "link_handler  -- cycle through xdg-open, webbrowser \n"
                    "                 and user-defined function\n"
                    "next          -- jump to next URL\n"
                    "open_queue    -- open all URLs in queue\n"
                    "open_queue_win-- open all URLs in queue in new window\n"
                    "open_url      -- open selected URL\n"
                    "palette       -- cycle through palettes\n"
                    "previous      -- jump to previous URL\n"
                    "quit          -- quit\n"
                    "reverse       -- reverse order URLs/context\n"
                    "shorten       -- toggle shorten highlighted URL\n"
                    "single        -- quit urlscan after opening a\n"
                    "                 single link\n"
                    "top           -- move to first list item\n"
                    "up            -- cursor up\n")
            self.top.base_widget.body = \
                urwid.ListBox(urwid.SimpleListWalker([urwid.Columns([(24, urwid.Text(help_men)),
                                                                     urwid.Text(docs)])]))
        else:
            self.top.base_widget.body = urwid.ListBox(self.items)
            self.top.base_widget.body.focus_position = self.focus_pos_saved
        self.help_menu = not self.help_menu

    def _search_key(self):
        """ / """
        if self.urls:
            self.search = True
            if self.compact is True:
                self._context()
        else:
            return
        self.no_matches = False
        self.search_string = ""
        # Reset the search highlighting
        self._search()
        footerwid = urwid.AttrMap(urwid.Text("Search: "), 'footer')
        self.top.base_widget.footer = footerwid
        self.items = self.items_orig
        self.top.base_widget.body = urwid.ListBox(self.items)

    def _digits(self):
        """ 0-9 """
        self.number += self.key
        try:
            if self.compact is False:
                self.top.base_widget.body.focus_position = \
                    self.items.index(self.items_com[max(int(self.number) - 1, 0)])
            else:
                self.top.base_widget.body.focus_position = \
                    self.items.index(self.items[max(int(self.number) - 1, 0)])
        except IndexError:
            self.number = self.number[:-1]
        except ValueError:
            pass
        self.top.base_widget.keypress(self.size, "")  # Trick urwid into redisplaying the cursor
        if self.number:
            self._footer_display(f"Selection: {self.number}", 1)

    def _clear_screen(self):
        """ Ctrl-l """
        self.draw_screen()

    def _down(self):
        """ j """
        self.top.base_widget.keypress(self.size, "down")

    def _up(self):
        """ k """
        self.top.base_widget.keypress(self.size, "up")

    def _top(self):
        """ g """
        # Goto top of the list
        self.top.base_widget.body.focus_position = 2 if self.compact is False else 0
        self.top.base_widget.keypress(self.size, "")  # Trick urwid into redisplaying the cursor

    def _bottom(self):
        """ G """
        # Goto bottom of the list
        self.top.base_widget.body.focus_position = len(self.items) - 1
        self.top.base_widget.keypress(self.size, "")  # Trick urwid into redisplaying the cursor

    def _selectable_positions(self):
        return [i for i, item in enumerate(self.items) if item.selectable()]

    def _next(self):
        """ J """
        current_position = self.top.base_widget.body.focus_position
        if current_position >= self._selectable_positions()[-1]:
            # Do not jump if focus is on or after the last selectable position
            return
        # Jump to the first selectable position after the currently focused position
        target_position = min(p for p in self._selectable_positions() if p > current_position)
        self.top.base_widget.body.focus_position = target_position
        self.top.base_widget.keypress(self.size, "")  # Trick urwid into redisplaying the cursor

    def _page_up(self):
        """ Ctrl-b """
        self.top.base_widget.keypress(self.size, "page up")
        self.top.base_widget.keypress(self.size, "")  # Trick urwid into redisplaying the cursor

    def _page_down(self):
        """ Ctrl-f """
        self.top.base_widget.keypress(self.size, "page down")
        self.top.base_widget.keypress(self.size, "")  # Trick urwid into redisplaying the cursor

    def _previous(self):
        """ K """
        current_position = self.top.base_widget.body.focus_position
        if current_position <= self._selectable_positions()[0]:
            # Do not jump if focus is on or before the first selectable position
            return
        # Jump to the first selectable position before the currently focused position
        target_position = max(p for p in self._selectable_positions() if p < current_position)
        self.top.base_widget.body.focus_position = target_position
        self.top.base_widget.keypress(self.size, "")  # Trick urwid into redisplaying the cursor

    def _shorten(self):
        """ s """
        # Toggle shortened URL for selected item
        fpo = self.top.base_widget.body.focus_position
        url_idx = len([i for i in self.items[:fpo + 1]
                       if isinstance(i, urwid.Columns)]) - 1
        if self.compact is False and fpo <= 1:
            return
        url = self.urls[url_idx]
        short = not "..." in self.items[fpo][1].label
        self.items[fpo][1].set_label(shorten_url(url, self.size[0], short))

    def _all_shorten(self):
        """ S """
        # Toggle all shortened URLs
        self.shorten = not self.shorten
        urls = iter(self.urls)
        for item in self.items:
            # Each Column has (Text, Button). Update the Button label
            if isinstance(item, urwid.Columns):
                item[1].set_label(shorten_url(next(urls),
                                              self.size[0],
                                              self.shorten))

    def _all_escape(self):
        """ u """
        # Toggle all escaped URLs
        self.unesc = not self.unesc
        self.urls, self.urls_unesc = self.urls_unesc, self.urls
        urls = iter(self.urls)
        for item in self.items:
            # Each Column has (Text, Button). Update the Button label
            if isinstance(item, urwid.Columns):
                item[1].set_label(shorten_url(next(urls),
                                              self.size[0],
                                              self.shorten))

    def _reverse(self):
        """ R """
        # Reverse items
        fpo = self.top.base_widget.body.focus_position
        if self.compact is True:
            self.items.reverse()
        else:
            rev = []
            for item in self.items:
                if isinstance(item, urwid.Divider):
                    rev.insert(0, item)
                elif isinstance(item, urwid.Text):
                    rev.insert(1, item)
                else:
                    rev.insert(2, item)
            self.items = rev
        self.top.base_widget.body = urwid.ListBox(self.items)
        self.top.base_widget.body.focus_position = self._cur_focus(fpo)

    def _context(self):
        """ c """
        # Show/hide context
        if self.search_string:
            # Reset search when toggling compact mode
            footerwid = urwid.AttrMap(urwid.Text(""), 'default')
            self.top.base_widget.footer = footerwid
            self.search_string = ""
            self.items = self.items_orig
        fpo = self.top.base_widget.body.focus_position
        self.items, self.items_com = self.items_com, self.items
        self.top.base_widget.body = urwid.ListBox(self.items)
        self.top.base_widget.body.focus_position = self._cur_focus(fpo)
        self.compact = not self.compact

    def _clipboard(self, pri=False):
        """ C """
        # Copy highlighted url to clipboard
        fpo = self.top.base_widget.body.focus_position
        url_idx = len([i for i in self.items[:fpo + 1]
                       if isinstance(i, urwid.Columns)]) - 1
        if self.compact is False and fpo <= 1:
            return
        url = self.urls[url_idx]
        cmds = COPY_COMMANDS_PRIMARY if pri else COPY_COMMANDS
        for cmd in cmds:
            try:
                subprocess.run(shlex.split(cmd),
                               check=False,
                               input=url.encode(sys.getdefaultencoding()),
                               stdout=subprocess.DEVNULL,
                               stderr=subprocess.DEVNULL)
                self._footer_display("Copied url to "
                                     f"{'primary' if pri is True else 'clipboard'} selection", 5)
            except OSError:
                continue
            if self.single is True:
                self._quit()
            break

    def _clipboard_pri(self):
        """ P """
        # Copy highlighted url to primary selection
        self._clipboard(pri=True)

    def _palette(self):
        """ p """
        # Loop through available palettes
        self.palette_idx += 1
        try:
            self.loop.screen.register_palette(self.palettes[self.palette_names[self.palette_idx]])
        except IndexError:
            self.loop.screen.register_palette(self.palettes[self.palette_names[0]])
            self.palette_idx = 0
        self.loop.screen.clear()

    def _config_create(self):
        """ --genconf """
        # Create ~/.config/urlscan/config.json if if doesn't exist
        if not exists(self.conf):
            os.makedirs(dirname(expanduser(self.conf)), exist_ok=True)
            keys = dict(zip(self.keys.keys(),
                            [i.__name__.strip('_') for i in self.keys.values()]))
            with open(expanduser(self.conf), 'w', encoding=sys.getdefaultencoding()) as pals:
                pals.writelines(json.dumps({"palettes": self.palettes, "keys": keys},
                                           indent=4))
            print("Created ~/.config/urlscan/config.json")
        else:
            print("~/.config/urlscan/config.json already exists")

    def _footer_display(self, text, time):
        """Display given text in the footer. Clears after <time> seconds

        """
        footerwid = urwid.AttrMap(urwid.Text(text), 'footer')
        self.top.base_widget.footer = footerwid
        self.loop.set_alarm_in(time, self._footer_callback)

    def _footer_callback(self, _loop, _data):
        """Callback for loop set_alarm_in after launching a URL or displaying a
        URL selection number, clearing the screen and clearing the footer
        loading message.

        """
        self.number = ""  # Clear URL selection number
        text = f"Search: {self.search_string}"
        if self.search_string:
            footer = 'search'
        else:
            footer = 'default'
            text = ""
        footerwid = urwid.AttrMap(urwid.Text(text), footer)
        self.top.base_widget.footer = footerwid
        self.draw_screen()

    def _cur_focus(self, fpo=0):
        # Return correct focus when toggling 'show context'
        if self.compact is False:
            idx = len([i for i in self.items_com[:fpo + 1]
                       if isinstance(i, urwid.Columns)]) - 1
            idx = max(idx, 0)
        elif self.compact is True:
            idx = [i for i in enumerate(self.items)
                   if isinstance(i[1], urwid.Columns)][fpo][0]
        return idx

    def _search(self):
        """ Search - search URLs and text.

        """
        text = f"Search: {self.search_string}"
        footerwid = urwid.AttrMap(urwid.Text(text), 'footer')
        self.top.base_widget.footer = footerwid
        search_items = []
        for grp in self.items_org:
            done = False
            for idx, item in enumerate(grp):
                if isinstance(item, urwid.Columns):
                    for col_idx, col in enumerate(item.contents):
                        if isinstance(col[0], urwid.AttrMap):
                            grp[idx][col_idx].set_label(splittext(col[0].base_widget.label,
                                                                  self.search_string,
                                                                  ''))
                            if self.search_string.lower() in col[0].base_widget.label.lower():
                                grp[idx][col_idx].set_label(splittext(col[0].base_widget.label,
                                                                      self.search_string,
                                                                      'search'))
                                done = True
                elif isinstance(item, urwid.Text):
                    grp[idx].set_text(splittext(item.text, self.search_string, ''))
                    if self.search_string.lower() in item.text.lower():
                        grp[idx].set_text(splittext(item.text, self.search_string, 'search'))
                        done = True
            if done is True:
                search_items.extend(grp)
        self.items = search_items
        self.top.base_widget.body = urwid.ListBox(self.items)
        if self.items:
            self.top.base_widget.body.focus_position = 2 if self.compact is False else 0
            # Trick urwid into redisplaying the cursor
            self.top.base_widget.keypress(self.size, "")
            self.no_matches = False
        else:
            self.no_matches = True
            footerwid = urwid.AttrMap(urwid.Text(text + "  No Matches"), 'footer')
            self.top.base_widget.footer = footerwid

    def draw_screen(self):
        """Render curses screen

        """
        self.tui.clear()
        canvas = self.top.base_widget.render(self.size, focus=True)
        self.tui.draw_screen(self.size, canvas)

    def _get_search(self):
        return lambda: self.search, lambda: self.enter

    def _link_handler(self):
        """Function to cycle through opening links via webbrowser module,
        xdg-open or custom expression passed with --run-safe or --run.

        """
        mode = self.link_open_modes.pop()
        self.link_open_modes.insert(0, mode)
        if self.nohelp is False:
            self.headerwid = urwid.AttrMap(urwid.Text(
                self.header.format(self.link_open_modes[0], len(self.queue))), 'header')
            self.top.base_widget.header = self.headerwid

    def mkbrowseto(self, url, mode=0, background=False):
        """Create the urwid callback function to open the web browser or call
        another function with the URL.

        Args: background (default False)
                If False, runs browser in a thread so it doesn't block the
                urwid event loop.

        """
        def browse(*args):  # pylint: disable=unused-argument
            # double ()() to ensure self.search evaluated at runtime, not when
            # browse() is _created_. [0] is self.search, [1] is self.enter
            # self.enter prevents opening URL when in search mode
            if self._get_search()[0]() is True:
                if self._get_search()[1]() is True:
                    self.search = False
                    self.enter = False
                return
            if self.link_open_modes[0] == "Web Browser" \
                    and os.environ.get('BROWSER') in TERMINAL_BROWSERS:
                # Text-mode browsers need control of the real terminal and
                # their own stdout, so suspend the urwid screen and run them
                # in the foreground.
                self.loop.stop()
                try:
                    subprocess.run([os.environ['BROWSER'], url], check=False)
                finally:
                    self.loop.start()
                    self.draw_screen()
                if self.single is True:
                    self._quit()
                return
            with redirect_output():
                if self.link_open_modes[0] == "Web Browser":
                    webbrowser.open(url, new=mode)
                elif self.link_open_modes[0] == "Xdg-Open":
                    subprocess.run(shlex.split(f'xdg-open "{url}"'), check=False)
                elif self.link_open_modes[0] == self.runsafe:
                    if self.pipe:
                        subprocess.run(shlex.split(self.runsafe),
                                       check=False,
                                       input=url.encode(sys.getdefaultencoding()))
                    else:
                        cmd = [i.format(url) for i in shlex.split(self.runsafe)]
                        subprocess.run(cmd, check=False)
                elif self.link_open_modes[0] == self.run and self.pipe:
                    subprocess.run(shlex.split(self.run),
                                   check=False,
                                   input=url.encode(sys.getdefaultencoding()))
                else:
                    subprocess.run(self.run.format(url), check=False, shell=True)

                if self.single is True:
                    self._quit()

        if background or os.environ.get('BROWSER') in TERMINAL_BROWSERS:
            return browse

        def browse_async(*args):  # pylint: disable=unused-argument
            # daemon=True so quitting urlscan doesn't wait for a still-open
            # browser to close.
            Thread(target=browse, args=args, daemon=True).start()
        return browse_async

    def process_urls(self, extractedurls, dedupe, shorten):
        """Process the 'extractedurls' and ready them for either the curses browser
        or non-interactive output

        Args: extractedurls
              dedupe - Remove duplicate URLs from list

        Returns: items - List of widgets for the ListBox
                 urls - List of all URLs

        """
        items = []
        urls = []
        for group, usedfirst, usedlast in extractedurls:
            items.append(urwid.Divider(div_char='-', top=1, bottom=1))
            if dedupe is True:
                # If no unique URLs exist, then skip the group completely
                if not [chunk for chunks in group for chunk in chunks
                        if chunk.url is not None and chunk.url not in urls]:
                    continue
            groupurls = []
            markup = []
            if not usedfirst and not self.whitespaceoff:
                markup.append(('msgtext:ellipses', '...\n'))
            for chunks in group:
                i = 0
                while i < len(chunks):
                    chunk = chunks[i]
                    i += 1
                    if chunk.url is None:
                        markup.append(('msgtext', chunk.markup))
                    else:
                        chunk.url = chunk.url.strip()
                        if (dedupe is True and chunk.url not in urls) \
                                or dedupe is False:
                            urls.append(chunk.url)
                            groupurls.append(chunk.url)
                        # Collect all immediately adjacent
                        # chunks with the same URL.
                        tmpmarkup = []
                        if chunk.markup:
                            tmpmarkup.append(('msgtext', chunk.markup))
                        while i < len(chunks) and \
                                (chunks[i].url if chunks[i].url is None
                                    else chunks[i].url.strip()) == chunk.url:
                            if chunks[i].markup:
                                tmpmarkup.append(chunks[i].markup)
                            i += 1
                        url_idx = urls.index(chunk.url) + 1 if dedupe is True else len(urls)
                        markup += [tmpmarkup or '<URL>',
                                   ('urlref:number:braces', ' ['),
                                   ('urlref:number', repr(url_idx)),
                                   ('urlref:number:braces', ']')]
                if not self.whitespaceoff:
                    markup += '\n'
            if not usedlast and not self.whitespaceoff:
                markup += [('msgtext:ellipses', '...\n\n')]
            items.append(urwid.Text(markup))

            i = len(urls) - len(groupurls)
            for url in groupurls:
                i += 1
                markup = [(6, urwid.Text([('urlref:number:braces', '['),
                                          ('urlref:number', repr(i)),
                                          ('urlref:number:braces', ']'),
                                          ' '])),
                          urwid.AttrMap(urwid.Button(shorten_url(url,
                                                                 self.width,
                                                                 shorten),
                                                     self.mkbrowseto(url),
                                                     user_data=url),
                                        'urlref:url', 'url:sel')]
                items.append(urwid.Columns(markup))

        return items, urls


@contextlib.contextmanager
def redirect_output():
    """
    A context manager to temporarily redirect stderr and stdout to devnull

    Usage:
        with redirect_output():
            webbrowser.open('https://google.com')

    """
    try:
        err = os.dup(sys.stderr.fileno())
        out = os.dup(sys.stdout.fileno())
        dest_file = open(os.devnull, 'w')
        os.dup2(dest_file.fileno(), sys.stderr.fileno())
        os.dup2(dest_file.fileno(), sys.stdout.fileno())
        yield
    finally:
        if err is not None:
            os.dup2(err, sys.stderr.fileno())
        if out is not None:
            os.dup2(out, sys.stdout.fileno())
        if dest_file is not None:
            dest_file.close()



#!/usr/bin/env python3
""" A simple urlview replacement that handles things like quoted-printable
properly.

"""
#
#   Copyright (C) 2006-2007 Daniel Burrows
#   Copyright (C) 2023 Scott Hansen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA

import argparse
import io
import os
import sys
from email import policy
from email.parser import BytesParser
from urlscan import urlchoose, urlscan
from urlscan._version import version


def parse_arguments():
    """Parse command line options.

    Returns: args

    """
    arg_parse = argparse.ArgumentParser(description="Parse and display URLs")
    arg_parse.add_argument('--genconf', '-g',
                           action='store_true', default=False,
                           help="Generate config file and exit.")
    arg_parse.add_argument('--compact', '-c',
                           action='store_true', default=False,
                           help="Don't display the context of each URL.")
    arg_parse.add_argument('--reverse', '-R', dest="reverse",
                           action='store_true', default=False,
                           help="Reverse order of displayed URLS/context")
    arg_parse.add_argument('--no-browser', '-n', dest="nobrowser",
                           action='store_true', default=False,
                           help="Pipe URLs to stdout")
    arg_parse.add_argument('--dedupe', '-d', dest="dedupe",
                           action='store_true', default=False,
                           help="Remove duplicate URLs from list")
    arg_parse.add_argument('--regex', '-E',
                           help="Alternate custom regex to be used for all "
                           "kinds of matching. "
                           r"For example: --regex 'https?://.+\.\w+'")
    arg_parse.add_argument('--run', '-r',
                           help="Alternate command to run on selected URL "
                           "instead of opening URL in browser. Use {} to "
                           "represent the URL value in the expression. "
                           "For example: --run 'echo {} | xclip -i'")
    arg_parse.add_argument('--run-safe', '-f', dest="runsafe",
                           help="Alternate command to run on selected URL "
                           "instead of opening URL in browser. Use {} to "
                           "represent the URL value in the expression. Safest "
                           "run option but uses `shell=False` which does not "
                           "allow use of shell features like | or >. Can use "
                           "with --pipe.")
    arg_parse.add_argument('--pipe', '-p', dest='pipe',
                           action='store_true', default=False,
                           help="Pipe URL into the command specified by --run or --run-safe")
    arg_parse.add_argument('--nohelp', '-H', dest='nohelp',
                           action='store_true', default=False,
                           help='Hide help menu by default')
    arg_parse.add_argument('--single', '-s', dest='single',
                           action='store_true', default=False,
                           help='Quit urlscan after opening/copying a single link.')
    arg_parse.add_argument('--width', '-w', dest='width',
                           type=int, default=0,
                           help='Set width to display')
    arg_parse.add_argument('--whitespace-off', '-W', dest='whitespaceoff',
                           action='store_true', default=False,
                           help="Don't display empty lines and ellipses.")
    arg_parse.add_argument('--headers', dest='headers',
                           action='store_true', default=False,
                           help='Scan certain message headers for URLs.')
    arg_parse.add_argument('--colors', '-C',
                           choices=['true', '256', '88', '16', '8', 'mono'],
                           default='16',
                           help='Set the terminal color palette (true=2**24, mono=1)')
    arg_parse.add_argument('--version', '-V', dest='version',
                           action='store_true', default=False,
                           help='Print urlscan version')
    arg_parse.add_argument('message', nargs='?', default=sys.stdin,
                           help="Filename of the message to parse")
    return arg_parse.parse_args()


def close_stdin():
    """This section closes out sys.stdin if necessary so as not to block curses
    keyboard inputs

    """
    if not os.isatty(0):
        try:
            fdesc = os.open('/dev/tty', os.O_RDONLY)
        except OSError:
            # This is most likely a non-interactive session, try to open
            # `stdin` directly
            fdesc = os.open('/dev/stdin', os.O_RDONLY)

        if fdesc < 0:
            sys.stderr.write('Unable to open an input tty.\n')
            sys.exit(-1)
        else:
            os.dup2(fdesc, 0)
            os.close(fdesc)


def process_input(fname):
    """Return the parsed text of stdin or the message. Accounts for possible
    file encoding differences.

        Args: fname - filename or sys.stdin
        Returns: mesg - EmailMessage object

    """
    if fname is sys.stdin:
        try:
            stdin_file = fname.buffer.read()
        except AttributeError:
            stdin_file = fname.read()
    else:
        stdin_file = None
    if stdin_file is not None:
        fobj = io.BytesIO(stdin_file)
    else:
        fobj = io.open(fname, mode='rb')
    f_keep = fobj
    mesg = BytesParser(policy=policy.default.clone(utf8=True)).parse(fobj)
    if 'From' not in mesg.keys() and 'Date' not in mesg.keys():
        # If it's not an email message, don't let the email parser
        # delete the first line. If it is, let the parser do its job so
        # we don't get mailto: links for all the To and From addresses
        fobj = _fix_first_line(f_keep)
        mesg = BytesParser(policy=policy.default.clone(utf8=True)).parse(fobj)
    try:
        fobj.close()
    except NameError:
        pass
    close_stdin()
    return mesg


def _fix_first_line(fline):
    """If the first line starts with http* or [ or other non-text characters,
    the URLs on that line will not be parsed by email.Parser. Add a blank line
    at the top of the file to ensure everything is read in a non-email file.

    """
    fline.seek(0)
    new = io.BytesIO()
    new.write(b"\n" + fline.read())
    fline.close()
    new.seek(0)
    return new


def main():
    """Entrypoint function for urlscan

    """
    args = parse_arguments()
    if args.version is True:
        print(version)
        return
    if args.genconf is True:
        urlchoose.URLChooser([], genconf=True)
        return
    msg = process_input(args.message)
    if args.nobrowser is False:
        tui = urlchoose.URLChooser(urlscan.msgurls(msg, regex=args.regex, headers=args.headers),
                                   compact=args.compact,
                                   reverse=args.reverse,
                                   nohelp=args.nohelp,
                                   dedupe=args.dedupe,
                                   run=args.run,
                                   runsafe=args.runsafe,
                                   single=args.single,
                                   width=args.width,
                                   whitespaceoff=args.whitespaceoff,
                                   pipe=args.pipe,
                                   colors=args.colors)
        tui.main()
    else:
        out = urlchoose.URLChooser(urlscan.msgurls(msg, regex=args.regex, headers=args.headers),
                                   dedupe=args.dedupe,
                                   reverse=args.reverse,
                                   shorten=False)
        if args.reverse is True:
            out.urls.reverse()
        print("\n".join(out.urls))


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
#   Copyright (C) 2006-2007 Daniel Burrows
#   Copyright (C) 2023 Scott Hansen
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

"""Contains the backend logic that scans messages for URLs and context."""

from html.parser import HTMLParser
import locale
import os
import re
from sys import getdefaultencoding


class Chunk:
    '''Represents a chunk of (marked-up) text that
    may or may not be linked to a URL.

    Attributes:
      markup - how this chunk will be rendered via urwid.
               This may be None if url is set, indicating
               that a URL footnote should be autogenerated.

      url    - the URL to which this text is linked, or None
               if no URL link is present.'''
    def __init__(self, markup, url):
        self.markup = markup
        self.url = url

    def __str__(self):
        return f'Chunk(markup = {repr(self.markup)}, url= {repr(self.url)})'

    def __repr__(self):
        return self.__str__()


def isheadertag(tag):
    """Determine if tag is a header """
    return len(tag) == 2 and tag[0] == 'h' and tag[1].isdigit()


class HTMLChunker(HTMLParser):
    """An HTMLParser that generates a sequence of lists of chunks.
    Each list represents a single paragraph."""

    def __init__(self):
        HTMLParser.__init__(self)

        # This attribute is the current output list.
        self.rval = []

        # If this attribute is True, the next chunk added will start a
        # new list.
        self.at_para_start = True
        self.trailing_space = False

        self.style_stack = [set()]
        self.anchor_stack = [None]
        self.list_stack = []
        # either 'ul' or 'ol' entries.

        # Ignore everything inside <style> and <script> elements
        self.in_style_or_script = False

    # Styled text uses the named attribute
    # msgtext:style1style2... where the styles
    # are always sorted in alphabetic order.
    # (if the text is in an anchor, substitute
    # "msgtext:anchor" for "msgtext")
    tag_styles = {'b': 'bold', 'i': 'italic'}

    ul_tags = ['*', '+', '-']

    def cur_url(self):
        return self.anchor_stack[-1]

    def add_chunk(self, chunk):
        if self.at_para_start:
            self.rval.append([])
        elif self.trailing_space:
            self.rval[-1].append(Chunk(' ', self.cur_url()))

        self.rval[-1].append(chunk)
        self.at_para_start = False
        self.trailing_space = False

    def end_para(self):
        if self.at_para_start:
            self.rval.append([])
        else:
            self.at_para_start = True
        self.trailing_space = False
        if self.list_stack:
            self.add_chunk(Chunk(' ' * 3 * len(self.list_stack),
                                 self.cur_url()))

    def end_list_para(self):
        if self.at_para_start:
            self.rval.append([])
        if self.list_stack:
            tag = self.list_stack[-1][0]
            if tag == 'ul':
                depth = len([t for t in self.list_stack if t[0] == tag])
                ul_tags = HTMLChunker.ul_tags
                chunk = Chunk(f"{ul_tags[depth % len(ul_tags)]}  ", self.cur_url())
            else:
                counter = self.list_stack[-1][1]
                self.list_stack[-1] = (tag, counter + 1)
                chunk = Chunk(f"{counter:2d}.", self.cur_url())
            self.add_chunk(chunk)
        else:
            self.end_para()

    def findattr(self, attrs, searchattr):
        for attr, val in attrs:
            if attr == searchattr:
                return val

        return None

    def handle_starttag(self, tag, attrs):
        if tag == 'a':
            self.anchor_stack.append(self.findattr(attrs, 'href'))
        elif tag in ('ul', 'ol'):
            self.list_stack.append((tag, 1))
            self.end_para()
        elif tag in HTMLChunker.tag_styles:
            self.style_stack.append(self.style_stack[-1] |
                                    set([HTMLChunker.tag_styles[tag]]))
        elif isheadertag(tag):
            self.style_stack.append(self.style_stack[-1] | set(['bold']))
        elif tag in ('p', 'br'):
            self.end_para()
        elif tag == 'img':
            # Since we expect HTML *email*, image links
            # should be external (naja?)
            alt = self.findattr(attrs, 'alt')
            if alt is None:
                alt = '[IMG]'
            src = self.findattr(attrs, 'src')
            if src is not None and not src.startswith(('http://', 'https://')):
                src = None

            if src is not None:
                self.anchor_stack.append(src)
                self.handle_data(alt)
                del self.anchor_stack[-1]
            else:
                self.handle_data(alt)
        elif tag == 'li':
            self.end_list_para()
        elif tag in ('style', 'script'):
            self.in_style_or_script = True

    def handle_startendtag(self, tag, attrs):
        if tag in set(['p', 'br', 'li', 'img']):
            self.handle_starttag(tag, attrs)

    def handle_endtag(self, tag):
        if tag == 'a':
            if len(self.anchor_stack) > 1:
                del self.anchor_stack[-1]
        elif tag in HTMLChunker.tag_styles:
            if len(self.style_stack) > 1:
                del self.style_stack[-1]
        elif tag in ('ul', 'ol'):
            if len(self.list_stack) > 0:
                del self.list_stack[-1]
            self.end_para()
        elif isheadertag(tag):
            if len(self.style_stack) > 1:
                del self.style_stack[-1]
            self.end_para()
        elif tag in ('style', 'script'):
            self.in_style_or_script = False

    def handle_data(self, data):
        if self.in_style_or_script:
            return
        future_trailing_space = False
        if data:
            if data[0].isspace():
                self.trailing_space = True
            if data[-1].isspace():
                future_trailing_space = True
        data = ' '.join(data.split())
        if self.anchor_stack[-1] is None:
            style = 'msgtext'
        else:
            style = 'anchor'
        if self.style_stack[-1]:
            stylelist = list(self.style_stack[-1])
            stylelist.sort()
            style = style + ':' + ''.join(stylelist)

        self.add_chunk(Chunk((style, data), self.cur_url()))
        self.trailing_space = future_trailing_space

    extrachars = {8212: "--",
                  8217: "'",
                  8220: "``",
                  8221: "''",
                  8230: "..."}

    def handle_charref(self, name):
        if name[0] == 'x':
            char = int(name[1:], 16)
        else:
            char = int(name)
        if char < 128:
            name = chr(char)
        elif char in HTMLChunker.extrachars:
            name = HTMLChunker.extrachars[char]
        else:
            name = f"&#{name};"
        self.handle_data(name)

    entities = {'nbsp': ' ',
                'lt': '<',
                'gt': '>',
                'amp': '&',
                'ldquo': '``',
                'rdquo': "''",
                'apos': "'"}

    def handle_entityref(self, name):
        if name in HTMLChunker.entities:
            self.handle_data(HTMLChunker.entities[name])
        else:
            # If you see a reference, it needs to be
            # added above.
            self.handle_data(f"&{name};")


URLINTERNALPATTERN = r'[\[\]{}()@\w/\\\-%?!&.=:;+,#~*]'
URLTRAILINGPATTERN = r'[{}(@\w/\-%&=+#$]'
HTTPURLPATTERN = (r'(?:(https?|file|ftps?)://' + URLINTERNALPATTERN +
                  r'*' + URLTRAILINGPATTERN + r')')
# Used to guess that blah.blah.blah.TLD is a URL.


def load_tlds():
    """Load all legal TLD extensions from assets

    """
    file = os.path.join(os.path.dirname(__file__),
                        'assets',
                        'tlds-alpha-by-domain.txt')
    with open(file, encoding=getdefaultencoding()) as fobj:
        return [elem for elem in fobj.read().lower().splitlines()[1:]
                if "--" not in elem]


TLDS = load_tlds()
GUESSEDURLPATTERN = (r'(?:[\w\-%]+(?:\.[\w\-%]+)*\.(?:' +
                     '|'.join(TLDS) + ')$)')
URLRE = re.compile(r'(?:<(?:URL:)?)?(' + HTTPURLPATTERN + '|' +
                   GUESSEDURLPATTERN +
                   r'|(?P<email>(mailto:)?[\w\-.]+@[\w\-.]*[\w\-]))>?',
                   flags=re.U)

# Poor man's test cases.
assert URLRE.match('<URL:http://linuxtoday.com>')
assert URLRE.match('http://linuxtoday.com')
assert re.compile(GUESSEDURLPATTERN).match('example.biz')
assert URLRE.match('example.biz')
assert URLRE.match('linuxtoday.com')
assert URLRE.match('master.wizard.edu')
assert URLRE.match('blah.bar.info')
assert URLRE.match('goodpr.org')
assert URLRE.match('http://github.com/firecat53/ürlscan')
assert URLRE.match('https://Schöne_Grüße.es/test')
assert URLRE.match('http://www.pantherhouse.com/newshelton/my-wife-thinks-i’m-a-swan/')
assert not URLRE.match('blah..org')
assert URLRE.match('http://www.testurl.zw')
assert URLRE.match('http://www.testurl.smile')
assert URLRE.match('testurl.smile.smile')
assert URLRE.match('testurl.biz.smile.zw')
assert not URLRE.match('example..biz')
assert not URLRE.match('blah.baz.obviouslynotarealdomain')
assert URLRE.match('http://[2a07:3500:11a0:320::22]:8080')


def parse_text_urls(mesg, regex=None):
    """Parse a block of text, splitting it into its url and non-url
    components."""

    rval = []

    loc = 0

    global URLRE

    if regex:
        URLRE = re.compile(regex)

    for match in URLRE.finditer(mesg):
        if loc < match.start():
            rval.append(Chunk(mesg[loc:match.start()], None))
        # Turn email addresses into mailto: links
        if regex:
            rval.append(Chunk(None, match.group(0)))
        else:
            email = match.group("email")
            if email and "mailto" not in email:
                mailto = f"mailto:{email}"
            else:
                mailto = match.group(1)
            rval.append(Chunk(None, mailto))
        loc = match.end()

    if loc < len(mesg):
        rval.append(Chunk(mesg[loc:], None))

    return rval


def extract_with_context(lst, pred, before_context, after_context):
    """Extract URL and context from a given chunk.

    """
    rval = []

    start = 0
    length = 0
    while start < len(lst):
        usedfirst = False
        usedlast = False
        # Extend to the next match.
        while start + length < len(lst) and length < before_context + 1 \
                and not pred(lst[start + length]):
            length += 1

        # Slide to the next match.
        while start + length < len(lst) and not pred(lst[start + length]):
            start += 1

        # If there was no next match, abort here (it's easier
        # to do this than to try to detect this case later).
        if start + length == len(lst):
            break

        # Now extend repeatedly until we can't find anything.
        while start + length < len(lst) and pred(lst[start + length]):
            extendlength = 1
            # Read in the 'after' context and see if it holds a URL.
            while extendlength < after_context + 1 and start + length + \
                    extendlength < len(lst) and \
                    not pred(lst[start + length + extendlength]):
                extendlength += 1
            length += extendlength
            if start + length < len(lst) and not pred(lst[start + length]):
                # Didn't find a matching line, so we either
                # hit the end or extended to after_context + 1..
                #
                # Now read in possible 'before' context
                # from the next URL; if we don't find one,
                # we discard the readahead.
                extendlength = 1
                while extendlength < before_context and start + length + \
                        extendlength < len(lst) and \
                        not pred(lst[start + length + extendlength]):
                    extendlength += 1
                if start + length + extendlength < len(lst) and \
                        pred(lst[start + length + extendlength]):
                    length += extendlength

        if length > 0 and start + length <= len(lst):
            if start == 0:
                usedfirst = True
            if start + length == len(lst):
                usedlast = True
            rval.append((lst[start:start + length], usedfirst, usedlast))

        start += length
        length = 0
    return rval


NLRE = re.compile('\r\n|\n|\r')


def extracturls(mesg, regex=None):
    """Given a text message, extract all the URLs found in the message, along
    with their surrounding context.  The output is a list of sequences of Chunk
    objects, corresponding to the contextual regions extracted from the string.

    """
    lines = NLRE.split(mesg)

    # The number of lines of context above to provide.
    # above_context = 1
    # The number of lines of context below to provide.
    # below_context = 1

    # Plan here is to first transform lines into the form
    # [line_fragments] where each fragment is a chunk as
    # seen by parse_text_urls.  Note that this means that
    # lines with more than one entry or one entry that's
    # a URL are the only lines containing URLs.

    linechunks = [parse_text_urls(i, regex=regex) for i in lines]

    return extract_with_context(linechunks,
                                lambda chunk: len(chunk) > 1 or
                                (len(chunk) == 1 and chunk[0].url is not None),
                                1, 1)


def extracthtmlurls(mesg):
    """Extract URLs with context from html type message. Similar to extracturls.

    """
    chunk = HTMLChunker()
    chunk.feed(mesg)
    chunk.close()
    # above_context = 1
    # below_context = 1

    def somechunkisurl(chunks):
        for chnk in chunks:
            if chnk.url is not None:
                return True
        return False

    return extract_with_context(chunk.rval, somechunkisurl, 1, 1)


def msgheaders(msg):
    """ Process email message headers for URLs

    Args: msg - email message object
    Returns: list

    """
    headers = ('Archived-At',
               'Link',
               'List-Archive',
               'List-ID',
               'List-Help',
               'List-Owner',
               'List-Post',
               'List-Subscribe',
               'List-Unsubscribe',
               'List-Unsubscribe-Post')
    res = []
    for hdr in headers:
        hdri = msg.get(hdr)
        if hdri:
            res.append(hdri)
    return res


def set_charset(message):
    """Get and/or set the message or message part charset. Try the
    content-charset or charset if it exists, or attempt to decode the message
    with a variety of charsets to find the correct one.

        Args: message - EmailMessage object
        Returns: message - EmailMessage object

    """
    if message.get_content_charset():
        return message
    if message.get_charset():
        return message
    enc_list = ['UTF-8', 'LATIN-1', 'iso8859-1', 'iso8859-2',
                'UTF-16', 'CP1252', 'CP720', 'CP437']
    locale.setlocale(locale.LC_ALL, '')
    code = locale.getpreferredencoding()
    if code not in enc_list:
        enc_list.insert(0, code)
    for enc in enc_list:
        try:
            message.as_bytes().decode(enc)
        except (UnicodeDecodeError, UnicodeError):
            continue
        else:
            try:
                message.set_param('charset', enc)
            except (KeyError, UnicodeEncodeError):
                # Try once to set correct encoding on the message part, then
                # continue without crashing if it fails
                continue
            break
        raise Exception("Encoding not detected.")
    return message


def msgurls(msg, urlidx=1, regex=None, headers=False):
    """Main entry function for urlscan.py

    """
    # Written as a generator so I can easily choose only
    # one subpart in the future (e.g., for
    # multipart/alternative).  Actually, I might even add
    # a browser for the message structure?
    if headers is True:
        for part in msgheaders(set_charset(msg)):
            for chunk in extracturls(part):
                urlidx += 1
                yield chunk
    msg = set_charset(msg)
    if msg.is_multipart():
        for part in msg.iter_parts():
            for chunk in msgurls(set_charset(part), urlidx, regex=regex):
                urlidx += 1
                yield chunk
    elif msg.get_content_type() == "text/plain":
        for chunk in extracturls(msg.get_content(), regex=regex):
            urlidx += 1
            yield chunk
    elif msg.get_content_type() == "text/html":
        for chunk in extracthtmlurls(msg.get_content()):
            urlidx += 1
            yield chunk
